#include <string.h>
#include <avr/io.h>
#include <avr/eeprom.h>

#include "vfd.h"
#include "vfd-put.h"
#include "ebook.h"

#include "ports.h"

#include "index.c"

/*****************************************************************/
// eBook einmal lesen: 167 Sekunden 
// eBook Daten: 4188215    (4090 kByte) in 167 Sekunden
// SPI   Daten: 3*4188215 (12270 kByte) in 132 Sekunden
// ==> Datendurchsatz EPROM: 24.5 kByte/s
// ==> Datendurchsatz SPI:     93 kByte/s
/*****************************************************************/

// last entry points to first 0xff, it is not a start of a book --> -1
#define INDEX_SIZE (sizeof (PGM_index) / sizeof (PGM_index[0]) -1)

eeprom_t eeprom EEMEM =
{
    .bright = 4,
    .speed  = 7,
    .show_vers = 1,
};

ebook_t ebook;

// read book address from index (PGM_index in index.c)
static inline uint32_t get_buch_addr (uint8_t idx);

void ebook_init() 
{
    vfd.bright = ebook.eeprom.bright;
    vfd.cursor_on = 0;
    vfd.cursor_blink = 0;
  
    vfd_clear();
}

// dir > 0: sucht aufwrts
// dir < 0: sucht abwrts
// sucht den |dir|-ten Fundort und
// stellt sich 1 hinter das gefundene Zeichen.
// return-Wert ist dir Anzahl der '\n' denen seek_byte begegnet
uint8_t seek_byte (const uint8_t target, const char dir)
{
    DDRD = 0x0;
    PORTD = 0x0;
    // Adjust Addresses and set appropriate CE to LOW (enable)
    prepare_addr (0);
 
    CLR (PORT_OE);
    char loops = (dir < 0) ? -dir : dir;
 
    uint8_t count = 0;
    
    while (1)
    {
        loops--;
   
        while (target != PIND)
        {
            if (PIND == '\n')
                count++;
            prepare_addr (dir);
        }
  
        if (PIND == '\n')
            count++;
  
        if (!loops)
            break;

        prepare_addr (dir);
    }
 
    SET (PORT_OE);
    prepare_addr (1);

    return count;
}

uint32_t get_buch_addr (uint8_t idx)
{
    return pgm_read_dword (& PGM_index[idx]);
}

// dir in {-1,0,1}
void seek_vers (char dir)
{
    uint8_t vers = ebook.nvers + dir;

    while (1)
    {
        seek_byte ('\n', (dir <= 0) ? dir-1 : dir);
        uint8_t data = read_byte();
  
        if ('_' == data || 0xff == data)
            vers = 0;
        else
            break;
    }
  
    if (0 == vers)
    {
        uint32_t addr = ebook.addr;
        uint8_t ncr = seek_byte ('_', -1);
        
        seek_kapitel (0);
        
        vers = 1;
        if (dir < 0)
            vers = ncr;
        
        ebook.addr = addr;
    }

    ebook.nvers = vers;
}

// dir in {-1,0,1}
void seek_kapitel (char dir)
{
    ebook.nvers = 1;
 
    if (dir <= 0)
        dir--;
  
    seek_byte ('_', dir);
    read_book_name();
}

// Sucht 
//  0: Anfang dieses Buchs
//  1: Anfang nchstes Buch
// -1: Anfang vorheriges Buch
// Liest name und stellt sich an den Anfang des ersten Verses
void seek_buch (char dir)
{
    ebook.nvers = 1;
    char i;
 
    for (i=0; i < INDEX_SIZE; i++)
        if (get_buch_addr (i) > ebook.addr)
            break;
 
    i += dir-1;
    if (i < 0)                  i += INDEX_SIZE;
    else if (i >= INDEX_SIZE)   i -= INDEX_SIZE;
 
    ebook.addr = get_buch_addr (i); 
 
    read_book_name();
}

// ebook.addr must be just behind a '_' (chapter start)
// read name of book
// set addr to start of first verse
void read_book_name()
{
    uint8_t data, *pname = ebook.name;
    uint8_t len = 0;

    do
    {
        prepare_addr (len ? 1 : 0);
        data = read_byte ();
  
        if (data == '\n')
            break;
  
        *pname++ = data;
    } while (++len < sizeof (ebook.name));
 
    *pname = '\0';
    while (*--pname != ' ');
    *pname++ = '\0';
 
    ebook.kapitel = pname;
 
    seek_byte ('\n', 1);
}

// liest ein byte von der momentanen Adresse aus dem richtigen EPROM
// vorher muss prepare_address() oder seek_byte() augerufen worden sein
uint8_t read_byte()
{
    uint8_t data;
 
    // Read data from EPROM
    DDRD  = 0x0;
    PORTD = 0x0;
 
    CLR (PORT_OE);
 
    CLR (PORT_OE); // nop
    CLR (PORT_OE); // nop

    data = PIND;

    SET (PORT_OE);
 
    return data;
}


///////////////////////////////////////////////////

void prepare_addr (char dir)
{
    uint32_t addr = ebook.addr;
 
    // saturate dir 
    if (dir < 0)  addr--;
    if (dir > 0)  addr++;
 
    ebook.addr = addr;
 
    // expand bits 20..21 of address to CE1...CE4 signals
    uint8_t ce = 1 << 0;
    if (addr & (1ul << 21))  ce = 1 << 2;
    if (addr & (1ul << 20))  ce = ce << 1;
 
    // set CEn bits 20..23
    addr |= (0xful << 20); 
 
    uint8_t foo;
 
    if (foo = ce & (1 << 0), foo)  addr &= ~(1ul << 20);
    if (foo = ce & (1 << 1), foo)  addr &= ~(1ul << 21);
    if (foo = ce & (1 << 2), foo)  addr &= ~(1ul << 22);
    if (foo = ce & (1 << 3), foo)  addr &= ~(1ul << 23);

    // unroll loop to speed up to the max
    // note that this as fast as if the bits were in order
 
#define OUT_BIT(n)                              \
    CLR (PORT_SER);                             \
    if (foo32 = addr & (1ul << n), foo32)       \
    SET (PORT_SER);                             \
    CLR (PORT_SCK);                             \
    SET (PORT_SCK)
 
    uint32_t foo32;

    // to understand the bit ordering refer to the schematic
    // IC 7
    OUT_BIT  (0); // 7
    OUT_BIT  (1); // 6
    OUT_BIT (10); // 5
    OUT_BIT  (2); // 4
    OUT_BIT (11); // 3
    OUT_BIT  (4); // 2
    OUT_BIT  (9); // 1
    OUT_BIT  (5); // 0

    // IC 6
    OUT_BIT  (23); // 7  CE4
    OUT_BIT  (22); // 6  CE3
    OUT_BIT  (17); // 5
    OUT_BIT  (21); // 4  CE2
    OUT_BIT   (3); // 3
    OUT_BIT  (12); // 2
    OUT_BIT  (15); // 1
    OUT_BIT  (20); // 0  CE1

    // IC 5
    OUT_BIT  (18); // 7
    OUT_BIT  (14); // 6
    OUT_BIT  (13); // 5
    OUT_BIT   (8); // 4
    OUT_BIT   (6); // 3
    OUT_BIT  (16); // 2
    OUT_BIT   (7); // 1
    OUT_BIT  (19); // 0
 
    // bring contents of 74*595 shift registers to output latches
    CLR (PORT_RCK);
    SET (PORT_RCK);
}
