#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>

#include "vfd.h"
#include "vfd-put.h"
#include "ebook.h"
#include "taster.h"
#include "timer.h"

#include "ports.h"

static void config();
static void ioinit(); 
static void print_vers (uint8_t mode, uint8_t align);

///////////////////////////////////////////////////////////////////////////////

// values for seekmode
enum
{
    SEEK_BUCH,
    SEEK_KAPITEL,
    SEEK_VERS
};

// values for the second argument of print_vers()
enum
{
    ALIGN,      // im Flu
    FLUSH_LEFT, // bei Auswahl Buch/Kapitel/Vers
    CLEAR,      // nach Verlassen von config()
    PREVIEW     // Blttern
};

// List of job functions that are to be called every 10ms

// job
typedef struct
{
    // job-Funktion
    void* job;

    // optional arg to pass
    uint8_t arg;
} job_t;

// the prototype of the job function may be one of
// <return-type> foo ()
// <return-type> foo (uint8_t)
// <return-type> foo (char)
// return value will be ignoren in the ISR
static const job_t jobs[] PROGMEM =
{
    {job_timer_10ms, 0}
    ,{get_taster, KEY_UP} 
    ,{get_taster, KEY_MITTE} 
    ,{get_taster, KEY_DOWN} 
};

// ISR will be called every 10ms and might dispatches to 
// one of the jobs.
// Every 10ms there starts a new frame.
// the 1st job called is jobs[0]
// the 2nd job calles is jobs[1] and so on.
// after 10ms the next frame restarts with jobs[0].
// This is to minimize the time that we spend in one ISR.
// For this simple program there is no need for this technique.
// This will, however, change for more sophisticated programs.
SIGNAL (SIG_OUTPUT_COMPARE1A)
{
    // count the number of IRQs in this frame, startin with 0.
    static uint8_t irq_num_10ms;

    // increment and overflow the IRQ count
    uint8_t int_num = 1+irq_num_10ms;
 
    // 10 Milli-Sekunden sind voll.
    if (int_num == IRQS_PER_10MS)
        int_num = 0;
  
    irq_num_10ms = int_num; 

    const int NUM_JOBS = sizeof (jobs) / sizeof (jobs[0]);
 
    // For IRQ number from 0 ... NUM_JOBS-1
    // there is a job to be executed
    if (int_num < NUM_JOBS)
    {
        // get job function to call
        uint16_t job = pgm_read_word (& jobs[int_num].job);

        // get the optional argument (a const)
        uint8_t  arg = pgm_read_byte (& jobs[int_num].arg);

        // execute the job function and pass the arg
        // note that there is some freedom in the callee's
        // prototype due to the AVR ABI.
        ((void(*)(uint8_t)) job) (arg);
    }
}

// Initialize ports to In, OUT, set PullUps, etc.
void ioinit (void) 
{
    // EPROM /OutputEnable
    SET (PORT_OE);  MAKE_OUT (PORT_OE);

    // VFD control
    SET (PORT_RESET); MAKE_OUT (PORT_RESET);
    CLR (PORT_E);     MAKE_OUT (PORT_E);
    CLR (PORT_RS);    MAKE_OUT (PORT_RS); 
    CLR (PORT_RW);    MAKE_OUT (PORT_RW);

    // VFD data lines to input
    VFD_IN_DB;

    // Taster are inputs with pullup
    MAKE_IN (PORT_TAST0); SET (PORT_TAST0);
    MAKE_IN (PORT_TAST1); SET (PORT_TAST1);
    MAKE_IN (PORT_TAST2); SET (PORT_TAST2);

    tasten[KEY_UP]   .mode = KM_REPEAT;
    tasten[KEY_MITTE].mode = KM_LONG;
    tasten[KEY_DOWN] .mode = KM_REPEAT;

    // Ports for Soft-SPI
    MAKE_OUT (PORT_SER);
    MAKE_OUT (PORT_SCK);  SET (PORT_SCK);
    MAKE_OUT (PORT_RCK);  SET (PORT_RCK);
}

// do some configuration
// - VFD's brightness
// - text shift speed
// - whether or not to display the actual "book chapter:verse"
void config()
{
    const uint8_t CFG_TIMEOUT = 5;
    
    enum
    {
        CFG_BRIGHT,
        CFG_SPEED,
        CFG_SHOW_VERSE,
        
        // sentinel
        CFG_MAX
    };
    
    uint8_t cfg = CFG_BRIGHT;
 
    while (1)
    {
        // refuel the config abort countdown 
        count.sec.config_timeout = CFG_TIMEOUT;

        // display the config target
  
        if (CFG_BRIGHT == cfg)  
            vfd_put ("%^Helligkeit%C%u", vfd.bright);

        if (CFG_SPEED == cfg)
            vfd_put ("%^Geschwindigkeit%C%u", ebook.eeprom.speed);

        if (CFG_SHOW_VERSE == cfg)
            vfd_put ("%^Vers-Nr anzeigen%C%S", 
                     ebook.eeprom.show_vers ? PSTR ("Ja") : PSTR ("Nein"));
    
        uint8_t tast;
  
        // exit this loop upon a key press
        do
        {
            tast = taster;

            // No key pressed for at least CFG_TIMEOUT seconds:
            // restore ebook.eeprom from EEPROM,
            // undo changes to VFD brightness
            // and return
            if (count.sec.config_timeout == 0)
            {  
                eeprom_read_block (& ebook.eeprom,
                                   & eeprom,
                                   sizeof (eeprom_t));
   
                vfd.bright = ebook.eeprom.bright;
                return;
            }
        } while (tast == NO_KEY);
  
        taster = NO_KEY;

        // do configuration according to key pressed
        // and selected config target
        if (CFG_BRIGHT == cfg)
        {
            uint8_t bright = ebook.eeprom.bright;
   
            if (tast == KEY_UP)     bright++;
            if (tast == KEY_DOWN)   bright--;

            bright &= 3;
            if (bright == 0)
                bright = 4;
    
            vfd.bright = ebook.eeprom.bright = bright;
        }
        if (CFG_SPEED == cfg)
        {
            if (tast == KEY_UP)      ebook.eeprom.speed++;
            if (tast == KEY_DOWN)    ebook.eeprom.speed--;
            ebook.eeprom.speed &= 15;
        }
        if (CFG_SHOW_VERSE == cfg)
        {
            if (tast == KEY_UP)      ebook.eeprom.show_vers = 1;
            if (tast == KEY_DOWN)    ebook.eeprom.show_vers = 0;
        }

        if (tast == KEY_MITTE)
            if (++cfg >= CFG_MAX)
                cfg = 0;

        // long key press:
        // store the .eeprom protion to EEPROM and return
        if (tast == KEY_MITTE_L)
        {
            eeprom_write_block (& ebook.eeprom,
                                & eeprom,
                                sizeof (eeprom_t));
            return;
        }
  
        wait_10ms (10);
    }
}

// print book/chapter/verse information to the bottom row
void print_vers (uint8_t mode, uint8_t align)
{
    // print to bottom leftmost position
    // and a preview of the respective verse to the top line
    if (PREVIEW == align)
    {
        prepare_addr (0);
  
        uint8_t i = 0;
        static uint8_t preview[16+1];
  
        while (i < sizeof (preview) - 1)
        {
            preview[i++] = read_byte();
            prepare_addr (1);
        } 
   
        vfd_put ("%^%s%C", preview);
    }

    // clear before output, align with text from eBook (after config())
    if (CLEAR == align)
        vfd_put ("%^%D");

    // align with text from eBook
    if (ALIGN == align)
        vfd_put ("%D ");

    // print to bottom leftmost position
    if (FLUSH_LEFT == align)
        vfd_put ("%C%* ", 40);
    // 
 
    uint8_t len_name, len_vers;
 
    len_name = vfd_put ("%s", ebook.name);
 
    if (mode < SEEK_KAPITEL)
        return;

    // avoid overflow if therer are long book names
    // instead of letting the cursor run out of sight
    // we overstrive long book names
    static char chap_n_verse[10];

    // print ebook.kapitel to chap_n_verse
    len_vers = vfd_put ("%@ %s", chap_n_verse, ebook.kapitel);

    if (mode >= SEEK_VERS)
        // append verse number to chap_n_verse
        len_vers += vfd_put ("%+:%u", chap_n_verse, ebook.nvers);

    // how much positions must we go left so that chapter and verse
    // completely fit into the visible area?
    char back = 0; 
    if (ALIGN != align)
    {
        back = VFD_VISIBLE_WIDTH - (len_name + len_vers); 
        if (back >= 0)
            back = 0;
    }

    // go back positions (back will be 0 or negative) and
    // print chapter and verse
    vfd_put ("%r%s", back, chap_n_verse); 
}


// out entry point after RESET and initialization of (module) globals
int main (void)
{
    ioinit ();

    // init Timer1 to make IRQs
    init_timer1();

    // enable IRQs globally
    sei();

    wait_10ms (30);

    // restore .eeprom portion of ebook from eeprom
    eeprom_read_block (& ebook.eeprom, & eeprom, sizeof (eeprom_t));
 
    // init VFD
    vfd_reset ();

    // init ebook
    ebook_init();

    uint8_t seekmode = SEEK_BUCH;

    // main loop
    while (1)
    {
        uint8_t tast = taster;

        // if there was a key press reset taster so that
        // get_taster() may put new values there
        if (NO_KEY != tast)
            taster = NO_KEY;

        // pressing the middle key
        // - shows info on the actual book/chapter/verse
        //   (dependent on seekmode) for 1.5s (150*10ms)
        // - after this time the display precess proceeds
        // - if the middle key is pressed again in that time
        //   then the seekmode is changed
        //   SEEK_BUCH --> SEEK_KAPITEL --> SEEK_VERS -> ...
        if (tast == KEY_MITTE)
        {
            if (count.ms10.key_timer)
            {
                if (++seekmode > SEEK_VERS)
                    seekmode = 0;
            }
            print_vers (seekmode, FLUSH_LEFT);
        }

        // a long press on the middle key goes to the
        // configuration menue
        if (tast == KEY_MITTE_L)
        {
            config();
            seek_vers (0);
   
            if (ebook.eeprom.show_vers)
                print_vers (SEEK_VERS, CLEAR);
            else
                vfd_clear ();
        }

        // key up will increase and key down will dedrease
        // book/chapter/verse (depends on seekmode)
        char dir = 0;
  
        if (tast == KEY_UP)    dir = 1;
        if (tast == KEY_DOWN)  dir = -1;
  
        if (dir)
        {
            if (seekmode == SEEK_VERS)     seek_vers    (dir);
            if (seekmode == SEEK_BUCH)     seek_buch    (dir);
            if (seekmode == SEEK_KAPITEL)  seek_kapitel (dir);
    
            print_vers (seekmode, PREVIEW);
        }

        // after pressing a key to display info or to
        // change book/chapter/verse there will be a delay of
        // 1.5 seconds before we start displaying and shifting (again)
        if (tast)
        {
            count.ms10.key_timer = 150;
  
            wait_10ms (10);
            continue;
        }

        if (count.ms10.key_timer)
            continue;

        // read byte from the EPROM memory
        prepare_addr (0);
  
        char c = read_byte();
  
        // end of eBook: seek start and rerun
        if (-1 == c)
        {
            seek_byte ('_', 1);
   
            // mimic the read of a '_'
            prepare_addr (-1);
            c = '_';
        }

        // next book or chapter
        if ('_' == c)
        {
            ebook.nvers = 0;
            prepare_addr (1);
            read_book_name();
   
            // mimic the read of a '\n'
            prepare_addr (-1);
            c = '\n';
        }
  
        // next verse
        if (c == '\n')
        {
            ebook.nvers++;
            if (ebook.eeprom.show_vers)
                print_vers (SEEK_VERS, ALIGN);
    
            // display '\n' as ' ' 
            c = ' '; 
        }

        // increment read address
        prepare_addr (1);
  
        // "%B%c" put c at first top position outside of visibility
        // "%C "  clear char at bottom left
        // "%<"   shift to left, c gets visible
        vfd_put ("%B%c%C %<", c);

        // slow down thing for humans
        count.ms10.ebook_delay = 2 * (15-ebook.eeprom.speed);
  
        while (count.ms10.ebook_delay)
            if (taster)  
                break;
    } // main loop
 
    // we will never come here (no return needed, is dead anyway)
}
