package de.gjlay.tools;

import java.io.*;
import java.net.URL;
/**
 * Is is a toolbox that
 * supplies some useful procedures.
 * @author Georg-Johann Lay
 */

public class FileTools extends Object
{
	private static final String NASTY_DOTS = "___NASTY_DOTS___/";
	private FileTools ()
		{
		}

	/**
	 * Read a file as one String from a Reader.
	 * @param reader	a {@link Reader} from which the String is to be
	 *					retrieved from.
	 * @param lineEnd	A String that is to be inserted after each
	 *					line of text. If <code>lineEnd</code> is
	 *					<code>null</code>, then the system dependent
	 *					line separator from the system property
	 *					<code>line.separator</code> is inserted.
	 * @return			The whole input read from <code>reader</code>
	 *					as one String.
	 * @throws IOException
	 * @see #getString(File,String)
	 * @see #getString(URL,String)
	 */
	public static String getString (Reader reader, String lineEnd)
		throws IOException
		{
			String line;
			LineNumberReader lineReader = new LineNumberReader (reader);

			StringBuffer result = new StringBuffer();

			if (null == lineEnd)
				lineEnd = System.getProperty ("line.separator");
			
			while (null != (line = lineReader.readLine ()))
			{
				result.append (line);
				result.append (lineEnd);
			}

			return result.toString();
		}
	
	/**
	 * Read a file as one String from a File.
	 * @param file		a {@link File} from which the String is to be
	 *					retrieved from.
	 * @param lineEnd	A String that is to be inserted after each
	 *					line of text. If <code>lineEnd</code> is
	 *					<code>null</code>, then the system dependant
	 *					line separator from the system property
	 *					<code>line.separator</code> is inserted.
	 * @return			The whole input read from <code>file</code>
	 *					as one String.
	 * @throws IOException
	 * @throws FileNotFoundException
	 * @see #getString(Reader,String)
	 * @see #getString(URL,String)
	 */
	public static String getString (File file, String lineEnd)
		throws IOException, FileNotFoundException
		{
			Reader reader = new FileReader (file);
			String result = getString (reader, lineEnd);

			reader.close();

			return result;
		}

	/**
	 * Read a file as one String from a URL.
	 * @param url		a {@link URL} from which the String is to be
	 *					retrieved from.
	 * @param lineEnd	A String that is to be inserted after each
	 *					line of text. If <code>lineEnd</code> is
	 *					<code>null</code>, then the system dependant
	 *					line separator from the system property
	 *					<code>line.separator</code> is inserted.
	 * @return			The whole input read from <code>url</code>
	 *					as one String.
	 * @throws IOException
	 * @see #getString(Reader,String)
	 * @see #getString(File,String)
	 */
	public static String getString (URL url, String lineEnd)
		throws IOException
		{
			String result;
			InputStream inStream = url.openStream();
			Reader reader = new InputStreamReader (inStream);
			result = getString (reader, lineEnd);
			inStream.close();

			return result;
		}

	/**
	 * Get a URL of a file that is located in a subdirectory of CLASSPATH
	 * @param path		the pathname of a resource
	 * @return			The URL of the resource or <code>null</code>,
	 *					if it does not exist.
	 */
	public static URL getUrlOfResource (String path)
		{
			URL url;

			path = StringTools.replace (path, "\\", "/");

			if (path.startsWith ("/"))
				path = path.substring (1);
			// Get rid of nasty ..'s that may occur in the path
			path = NASTY_DOTS + path;

			try
			{
				// resolve ..'s
				path = new File (path).getCanonicalPath();
			}
			catch (IOException ex)
			{
				return null;
			}

			int pos = path.indexOf (NASTY_DOTS) + NASTY_DOTS.length();
			path = path.substring (pos);
			// ok, path is ..'s free
			// 
			// search in CLASSPATH
			url = FileTools.class.getClassLoader().getResource (path);

			return url;
		}
	
	/**
	 * Write a String as a new File. Generate parent directories if required.
	 * @param fileName	the file name of the file
	 * @param str       the string that is to be written
	 * @throws IOException
	 * @see #writeStringAsFile(File,String)
	 */
	public static void writeStringAsFile (String fileName, String str)
		throws IOException
		{
			File outFile = new File (fileName);
			File parent  =  outFile.getParentFile();

			if (null != parent)
				parent.mkdirs();

			try
			{
				outFile.createNewFile();
				writeStringAsFile (outFile, str);
			}
			catch (IOException ex)
			{
				throw new IOException (ex.getMessage()
									   + ": " + outFile.toString());
			}
		}

	/**
	 * Write a string to an existing file
	 * @param outFile	the file where <code>str</code> is to be written to,
	 *					must exist
	 * @param str       the string
	 * @throws IOException
	 * @see #writeStringAsFile(String,String)
	 */
	public static void writeStringAsFile (File outFile, String str)
		throws IOException
		{
			FileWriter writer = new FileWriter (outFile);

			writer.write (str, 0, str.length());
			writer.close();
		}

	/**
	 * Get a {@link Reader} that reads from a file or from a resource located
	 * in a subdir of CLASSPATH
	 * @param fileName	the file name of the file. If <code>fileName</code>
	 *			starts with "<code>res:</code>",
	 *			then {@link #getReader(String)}
	 *			looks for a resource located in a subdir of CLASSPATH.
	 * @return	The reader
	 * @throws IOException
	 * @throws FileNotFoundException
	 */
	public static Reader getReader (String fileName)
		throws IOException, FileNotFoundException
		{
			Reader reader;
			
			if (fileName.startsWith ("res:"))
			{
				fileName = fileName.substring (4);
				URL fileUrl = getUrlOfResource (fileName);

				if (null == fileUrl)
				{
					throw new FileNotFoundException (
						"Resource '" + fileName
						+ "' not found in CLASSPATH");
				}
				
				InputStream inStream = fileUrl.openStream();
				
				reader = new InputStreamReader (inStream);
			}
			else
			{
				File file = new File (fileName);
				reader = new FileReader (file);
			}

			return reader;
		}
}
