package de.gjlay.tools;

import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import java.io.*;
import java.net.URL;
/**
 * <code>StringTools</code> is a toolbox that
 * supplies some useful procedures.
 * @author Georg-Johann Lay
 */

public class StringTools extends Object
{
	private static StringBuffer BLANKS;

	static
		{
			BLANKS = new StringBuffer();
			BLANKS.append ("                                        ");
			BLANKS.append ("                                        ");
			BLANKS.append ("                                        ");
		}
	
	private StringTools ()
		{
		}

	/**
	 * Trim and unquote a string: remove surrounding double quotes (&quot;).
	 * @param str		the string to unquote
	 * @return			an unquoted version of <code>str</code>.
	 *					If <code>str</code> is not enclosed in double
	 *					quotes, then the original string is returned.
	 * @see #fromQuotedCString(String, String, String)
	 */
	public static String unquote (String str)
		{
			str = str.trim();

			if (str.length() >= 2)
				if (str.startsWith ("\""))
					if (str.endsWith ("\""))
						str = str.substring (1, str.length()-1);

			return str;
		}

	/**
	 * Unquote a C string and replace C escape characters.
	 * Replacements:<br>
	 * <table>
	 * <tr><td><b><u>char sequence</u></b></td><td><b><u>replacement</u></b></td></tr>
	 * <tr><td><code>\n</code></td><td><code>back_n</code></td></tr>
	 * <tr><td><code>\t</code></td><td><code>back_t</code></td></tr>
	 * <tr><td><code>\\</code></td><td><code>\</code></td></tr>
	 * <tr><td><code>%%</code></td><td><code>%</code></td></tr>
	 * </table>
	 * @param str		the C string to unquote
	 * @param back_n	replacor string for '\n'
	 * @param back_t	replacor string for '\t'
	 * @return			the unqouted version of <code>str</code>.
	 *					Before doing the above replacements,
	 *					{@link #unquote} is executed on the string.
	 * @see #unquote(String)
	 */
	public static String fromQuotedCString (String str, String back_n, String back_t)
		{
			str = unquote (str);
			str = replace (str, "\\\"", "\"");
			str = replace (str, "\\n", back_n);
			str = replace (str, "\\t", back_t);
			str = replace (str, "\\\\", "\\");
			str = replace (str, "%%", "%");

			return str;
		}

	/**
	 * TeXify a string.<br>
	 * Escape characters that are special to TeX.
	 * 
	 * <table>
	 * <tr><td><b><u>char</u></b></td><td><b><u>replacement</u></b></td></tr>
	 * <tr><td><code>`</code> (back quote)</td><td><code>'</code> (single quote)</td></tr>
	 * <tr><td><code>\</code></td><td><code>{\textbackslash}</code></td></tr>
	 * <tr><td><code>&lt;</code></td><td><code>{\textless}</code></td></tr>
	 * <tr><td><code>&gt;</code></td><td><code>{\textgreater}</code></td></tr>
	 * <tr><td><code>{</code></td><td><code>\{</code></td></tr>
	 * <tr><td><code>}</code></td><td><code>\}</code></td></tr>
	 * <tr><td><code>#</code></td><td><code>\#</code></td></tr>
	 * <tr><td><code>_</code></td><td><code>\_</code></td></tr>
	 * <tr><td><code>%</code></td><td><code>\%</code></td></tr>
	 * <tr><td><code>&amp;</code></td><td><code>\&amp;</code></td></tr>
	 * <tr><td><code>$</code></td><td><code>\$</code></td></tr>
	 * <tr><td><code>^</code></td><td><code>\^{}</code></td></tr>
	 * <tr><td><code>~</code></td><td><code>\~{}</code></td></tr>
	 * <tr><td><code>--</code></td><td><code>-{}-</code></td></tr>
	 * <tr><td><code>|</code></td><td><code>{\textbar}</code></td></tr>
	 * <table>
	 * @param		str		string to be texified
	 * @return				a texified version of <code>str</code>.
	 */
	public static String toTexString (String str)
		{
			str = replace (str, "`", "'");
			str = replace (str, "\\", "\\\\");
			str = replace (str, "{", "\\{");
			str = replace (str, "}", "\\}");
			str = replace (str, "#", "\\#");
			str = replace (str, "_", "\\_");
			str = replace (str, "%", "\\%");
//			str = replace (str, "&", "{\\brezel}");
			str = replace (str, "&", "\\&");
			str = replace (str, "$", "\\$");
			str = replace (str, "<", "{\\textless}");
			str = replace (str, ">", "{\\textgreater}");
			str = replace (str, "|", "{\\textbar}");
			str = replace (str, "\\\\", "{\\textbackslash}");

			str = replace (str, "^", "\\^{}");
			str = replace (str, "~", "\\~{}");
			str = replace (str, "--", "-{}-");
			
			return str;
		}

	/**
	 * Replace all occurences of a substring with a new string.
	 * @param str		the target
	 * @param oldPat	the substring that is to replaced
	 * @param newPat	the replacement for all occurences
	 *					of <code>oldPat</code>
	 * @return			<code>str</code> with all occurences of
	 *					<code>oldPat</code> replaced by <code>newPat</code>.
	 *					Replacing is done from left to right. If a
	 *					replacement has been done, then <code>replace</code>
	 *					proceeds after the newly inserted <code>newPat</code>.
	 *					<code>newPat</code> may be <code>null</code>, in which
	 *					case it is treated like the empty string.
	 */					
	public static String replace (String str, String oldPat, String newPat)
		{
			int pos = 0;

			if (null == newPat)
				newPat = "";

			StringBuffer buf = new StringBuffer (str);
			
			while ((pos = buf.indexOf (oldPat, pos)) >= 0)
			{
				buf.replace (pos, pos+oldPat.length(), newPat);
				pos += newPat.length();
			}
			
			return buf.toString();
		}


	/**
	 * Replace specified characters by a new string.
	 * @param str		the target
	 * @param chars  	the string that contains the chars that are to
	 *                  be replaced
	 * @param newPat	the replacement for these characters
	 * @return			<code>str</code> with all occurences of
	 *					any char that appears in <code>chars</code>
	 *					replaced by <code>newPat</code>.
	 *					Replacing is done from left to right. If a
	 *					replacement has been done, then
	 *					<code>replaceChars</code> proceeds after the
	 *					newly inserted <code>newPat</code>.
	 */					
	public static String replaceChars (String str, String chars, String newPat)
		{
			if (null == newPat)
				newPat = "";

			StringBuffer buf = new StringBuffer (str);
			
			for (int index = 0; index < chars.length(); index++)
			{
				String oldPat = chars.substring (index, index+1);

				int pos = 0;
				while ((pos = buf.indexOf (oldPat, pos)) >= 0)
				{
					buf.replace (pos, pos+1, newPat);
					pos += newPat.length();
				}
			}
			
			return buf.toString();
		}

	/**
	 * Remove all trainling chars that occor in <code>chars</code>.
	 * @param str		the target
	 * @param chars  	the string that contains the chars that are to
	 *                  be removed
	 * @return			<code>str</code> with all trailing chars that
	 *					occur in <code>chars</code> removed.
	 */					
	public static String removeTrailingChars (String str, String chars)
		{
			StringBuffer buf = new StringBuffer (str);

			int len;
			
			while ((len = buf.length()) > 0)
			{
				char ch = buf.charAt (len-1);

				if (chars.indexOf (ch) >= 0)
					buf.setLength (len-1);
				else
					break;
			}

			return buf.toString();
		}

	/**
	 * Count all occurences of a char.
	 * @param str		the target
	 * @param c			the char whose occurences are to be counted
	 * @return			the number of occurences of
	 *					<code>c</code> in <code>str</code>.
	 */
	public static int getCharCount (String str, char c)
		{
			int count = 0;
			
			for (int index = 0; index < str.length(); index++)
				if (str.charAt (index) == c)
					count++;

			return count;
		}

	/**
	 * Convert a string to an integer.
	 * @param str		the string to be converted to an <code>int</code>
	 * @param value		a default value
	 * @return			<code>str</code> represented as
	 *					<code>int</code>. If the conversion cannot
	 *					be done for some reason, <code>value</code>
	 *					is returned.
	 */					
	public static int toInt (String str, int value)
		{
			try
			{
				value = Integer.parseInt (str);
			}
			catch (Exception e) {}

			return value;
		}

	/**
	 * Remove every char that is not a letter nor a whitespace.
	 * @param str	the target
	 * @return		a string with every char removed that is not
	 *				a letter nor a whitespace.
	 * @see Character#isLetter(char)
	 * @see Character#isWhitespace(char)
	 */
	public static String removeSpecialChars (String str)
		{
			int len = str.length();
			StringBuffer buf = new StringBuffer();

			for (int i=0; i<len; i++)
				{
					char c = str.charAt (i);

					if (Character.isLetter (c))
						buf.append (c);
					else
						if (Character.isWhitespace (c))
							buf.append (c);
				}

			return buf.toString();
		}

	/**
	 * Get a String of blanks of length <code>num</code> &lt;= 120.
	 */
	public static String getBlanks (int num)
		{
			return (num > 0)
				? BLANKS.substring (0, num)
				: "";
		}

	/**
	 * Get a Vector of Strings from a comma separated list.
	 * @param str	a list of comma separated tokens
	 * @return		a Vector of trimmed strings, one element for each token.
	 */
	public static Collection getCollection (String str)
		{
			return getCollection (str, Vector.class);
		}
			
	/**
	 * Get a Collection of Strings from a comma separated list.
	 * @param str	a list of comma separated tokens
	 * @param clazz	a Class that implements Collection
	 * @return		a collection of trimmed strings,
	 *              one element for each token.
	 *              You can order the tokens lexicographically
	 *              by using <code>TreeSet.class</code>, e.g.
	 *              To get the tokens in the order they occur,
	 *              pass <code>Vector.class</code>.
	 */
	public static Collection getCollection (String str, Class clazz)
		{
			Collection collect = null;
			
			try
			{
				collect = (Collection) clazz.newInstance();
			}
			catch (Exception ex)
			{
				ex.printStackTrace();
			}

			StringTokenizer tokens = new StringTokenizer (str, ",");
			
			while (tokens.hasMoreTokens())
			{
				collect.add (tokens.nextToken().trim());
			}

			return collect;
		}

	/**
	 * Convert a string to a Long.
	 * @param str		the string to be converted to a <code>Long</code>
	 * @return			<code>str</code> represented as
	 *					<code>Long</code>. If <code>str</code>
	 *					starts with a "0x", then
	 *					it is interpreted as a hex number.
	 */					
	public static Long toLong (String strVal)
		{
			int radix = 10;

			if (strVal.startsWith ("0x"))
			{
				strVal = strVal.substring (2);
				radix = 16;
			}

			Long val = null;
			
			val = new Long (Long.parseLong (strVal, radix));

			return val;
		}
		
}
