#ifndef _VFD_PUT_H_
#define _VFD_PUT_H_

#include <inttypes.h>
#include <avr/pgmspace.h>

// Clear VFD and reset corsor(s)
extern void vfd_clear();

/*
  Just for convenience to spare typing PSTR every time.
  This macro is very special in the way that it eats the superfluous
  komma if there are no arguments following x (mind the blanks!).
  So you may type
  
      vfd_put ("#^%s", "Hallo Welt!");
      
  which resolves to
  
      vfd_put_PSTR (PSTR ("#^%s"), "Hallo Welt!");
      
  while
  
      vfd_put ("#^Hallo Welt!");
      
  will resolve to
  
      vfd_put_PSTR (PSTR ("#^Hallo Welt!"));
*/      
#define vfd_put(x,y...)   vfd_put_PSTR (PSTR(x) , ##y)

/*
  The very VFD put function. The format string must reside in the flash.
  Format strings from RAM are very special and are not supported.
  
  This function works like the familiar printf-like functions:
  It gets a format string that it scans and from which it determines
  how to interpret and display the subsequent arguments.

  There are two groups of directives:
  %-directives that make output and are simlar to printf %-codes
  %-directives that are used to set the cursor and clear the display etc.

  All cursor positioning take place relative to the currently
  visible output window. Every line of the VFD has a length of
  40 characters from which only 16 are visible. As the display
  (and thereby the visible area) is shifted, new characters
  come into sight and others vanish. However, a %A will always position
  the corsor to the top left of the visible area.

  Actually therer are two cursors, the %0 for the top line and
  the %1 for the bottom line. So printing "%^Hallo%1Welt%0!" will
  clear the display and print

     Hallo!
     Welt

  The return value is the number of printed characters.
  
  Here are the directives:   
     
  %% :  %
  %s :  output string <ARG> from RAM
  %S :  output string <ARG> from Flash
  %c :  output char <ARG>
  %u :  output the unsigned int <ARG> as decimal number
  %d :  output the signed int <ARG> as decimal number
  %*c:  print <ARG> copies of character c
  %@ :  subsequent output will go to string <ARG>: sprintf functionality
  %+ :  subsequent output will be appended to string <ARG>
    
  %^ :  clear display, cursors home
  %p :  set cursor to relative (with respect to visible window) position <ARG>
  %r :  move cursor <ARG> positions left or right
  %0 :  subsequent output goes to top line
  %1 :  subsequent output goes to bottom line

  %A :  set cursor top left
  %B :  set cursor top right, first hidden position
  %C :  set cursor botton left
  %D :  set cursor bottom right, first hidden position

        The directives %A..%D are yust introduced for convenience.
        They are equivalent to %p with <ARG> 
        set to 0, 16, 0x40 resp. 0x40+16.
        Every call to vfd_put will resume printing at the cursor position
        that remain after the last call of vfd_put.

  %< :  shift left Display.
        Characters at %A and %C will be shifted out and %B and %D come
        into sight.

  %2 :  For internal use only, do not print to this channel. This channel
        will be used implicitly when printing/appending to a string with
        directives %@ and %+
        
  %3 :  subsequent output goes to channel %3.
        vfd-put.c does not put characters directly to the VFD.
        Instead, it uses putter functions from the array 'vfd_putters[]',
        e.g. output to the top line (%0) will be printed
        via vfd_putters[0]. If you want to hook in your own putter
        then yust increase the number of putters by adjusting
        PUTTER_CHANNELS below and set vfd_putters[n] to be the putter to
        be taken after a %n directive.
        Note that %3..%9 are neither used nor initialized by vfd-put.c.
        To be explicit, to set up channel %3 write

        // PUTTER_CHANNELS below must have been defined to be at least 4

        // my own put function 
        void my_putc (char c)
        {
           // do something (or nothing)
        }

        and in your code, before accessing %3

           // make channel %3 use my_putc
           vfd_putters[3] = my_putc;

           vfd_put ("%3Hallo Welt!");

        Putting to channel %2 or higher does not affect the cursors.
        
        vfd_putters[] can hold PUTTER_CHANNELS entries, so you can use
        channels %3..%<PUTTER_CHANNELS-1> for your purposes.
        %0, %1 and %2 are used by vfd-put.c. However, you may also replace
        these putters in order to meet your requirements.
*/
extern uint8_t vfd_put_PSTR (const char * fmt, ...);

/*
  The array of putter functions.
  Each entry is a pointer to a function whose prototype is
   
     void foo (char);
     
  Entries 0..2 are initialized by vfd-put.c upon system startup.
  If you want to use an output channel of 3 or higher you MUST
  initialize the corresponging entry of vfd_putters[] to be the
  desired function.
  Make sure that PUTTER_CHANNELS is defined properly. Using channel %n
  needs PUTTER_CHANNELS to be n+1 or higher.
*/
#define PUTTER_CHANNELS 3 
extern void (*vfd_putters[PUTTER_CHANNELS])(char);

#endif // _VFD_PUT_H_

